# Iowa::Client encapsulates the methods necessary to make a connection
# to an Iowa servlet, receive the response back from the servlet, and to
# deliver that response.

require 'socket'
require 'iowa/Request'

module Iowa

	class Client

		# When an Iowa::Client is created, it takes as its argument a
		# socket definition.  It uses that definition to establish either
		# a TCPSocket or a UNIXSocket to the servlet.
		#
		# The socket defenition can be either in the form of:
		#   hostname:port
		# or it can be a filesystem path.  If one uses the hostname:port
		# form, then a TCPSocket will be used.  Otherwise, a UNIXSocket
		# will be used.
		
		def initialize(sp)
			@start_time = Time.now
			socket_path = sp.dup
			socket_path.untaint
			tcp_pattern = Regexp.new('^([^:]*):(\d+)')
			tcp_match = tcp_pattern.match(socket_path)
			if tcp_match
				# Connect to a TCP socket.
				socket_host = tcp_match[1]
				socket_port = tcp_match[2]
				TCPSocket.do_not_reverse_lookup = true
				@socket = TCPSocket.new(socket_host,socket_port)
			else
				@socket = UNIXSocket.new(socket_path)
			end
		end

		# One calls initiate() to send the request off to the servlet.
		# The method wants to receive an Apache::Request object.
		# If it receives one, it will construct the Iowa::Request object
		# that is used to encapsulate the request to the servlet using
		# the provided Apache::Request object.
		
		def initiate(request=nil)
			ir = Iowa::Request.new(request)
			m = Marshal.dump(ir)
			@socket.write(m)
			@socket.flush
			@socket.shutdown(1)
		end

		# This receives the response back from the servlet.
		
		def receive
			m = ''
			@request = ''
			while (recv = @socket.recv(1024)) != ""
			  m << recv
			end
			@socket.close
			@socket = nil
			#@request = Marshal.load(m)
			@request = Marshal.load(m)
			@request.content_type = 'text/plain' unless @request.content_type.to_s != ''
		end

		# Prints/sends the HTTP headers for the response.
		
		def printHeaders(ios=nil)
			@request.calculate_runtime = false unless @request.content_type == 'text/html'
			calculate_runtime if @request.calculate_runtime
			if ENV['MOD_RUBY']
				r = Apache.request
				r.content_type = @request.content_type
				r.status_line = @request.status_line if @request.status_line
				@request.headers_out.each do |key,value|
					r.headers_out.set(key,value.to_s)
				end
				r.headers_out.set('Content-Length',@request.content.length.to_s)
				r.send_http_header
			elsif ios
				@request.headers_out['Date'] ||= Time.now.strftime('%a, %d %b %Y %H:%M:%S %Z')
				ios.print "Date: #{@request.headers_out['Date']}\n"

				ex = {'Date' => true, 'Server' => true, 'Connection' => true, 'Content-Type' => true}
				unless @request.headers_out['Content-Length']
					ex['Content-Length'] = true
					ios.print "Content-Length: #{@request.content.length}\n"
				end
				@request.headers_out.each do |key, value|
					next if ex.has_key? key
					ios.print "#{key}: #{value}\n"
				end
				ios.print "Content-Type: #{@request.content_type}\n"
				ios.print "\n"
			else
				@request.headers_out['Date'] ||= Time.now.strftime('%a, %d %b %Y %H:%M:%S %Z')
				@request.headers_out['Server'] ||= ENV['SERVER_SOFTWARE']
				puts "Date: #{@request.headers_out['Date']}"
				puts "Content-Type: #{@request.content_type}"
				ex = {'Date' => true, 'Server' => true, 'Connection' => true, 'Content-Type' => true}
				if @request.headers_out['Content-Length']
					ex['Content-Length'] = true
				else
					puts "Content-Length: #{@request.content.length}"
				end
				@request.headers_out.each do |key, value|
					next if ex.has_key? k
					puts "#{key}: #{value}"
				end
				puts ""
			end
		end

		# Deliver the content of the request.
		
		def printBody(ios=nil)
			if ios
				ios.print @request.content.to_s
			else
				puts @request.content.to_s
			end
		end

		def clearOld
			@socket = nil
			@request = nil
		end
		
		# Deliver both the header and the content of the request, then clear
		# the client.
		
		def print(ios = nil)
			printHeaders(ios)
			printBody(ios)
			clearOld
		end
		
		def calculate_runtime
			@runtime = Time.now.to_f - @start_time.to_f
			@request.content << "\n<!-- #{@runtime} -->\n"
		end
		
		def handle_request(args)
			initiate(args[:request])
			receive
			print(args[:ios])
		end
	end
end
