require 'iowa/JSON-lexer'
module Iowa

	# Provides high level functionality for all component subclasses.
	# Each "page" in an Iowa app is a Component subclass.

	class Component < Element

	  # Returns an empty hash.

		def Component.bindings
			{}
		end

	  # Returns the template for the component.

		def Component.template
			Iowa.app.templateForComponent(self.name)
		end

	  # Finds and returns the named component.  Other methods calling this
	  # one with different values for session and prevPage could induce some
	  # interesting (or bizarre) behaviors.  See pageNamed(name) below for
	  # the standard invocation.

		def Component.pageNamed(name, session, prevPage=nil)
			Iowa.app.reloadModified(name,true)
			klass = ElementClasses[name.to_s.downcase]
			if(klass && klass.ancestors.include?(Component))
				page = klass.new("page", {}, {}, nil)
				page.prevPage = prevPage
				page.session = session
				page
			else
				el = ''
				ElementClasses.each_pair do |k,v|
					el << "#{k} == #{v}<br/>\n"
				end
				raise "#{name}: no such page in #{el}"
			end
		end

	  # Provides a method to define component attributes which another
	  # component can bind to.
	
		def Component.attr_binding(*symbols)
			symbols.each do |symbol|
				class_eval %{
					def #{symbol}
						if @bindings.has_key?('#{symbol}')
							@bindings['#{symbol}'].get(@parent)
						else
							nil
						end
					end
					def #{symbol}=(val)
						if @bindings.has_key?('#{symbol}')
							@bindings['#{symbol}'].set(@parent, val)
						else
							@bindings['#{symbol}'] = LiteralAssociation.new(val)
						end
					end
				}
			end
		end
	
	  # Standard invocation of Component.pageNamed.  Returns the component
	  # corresponding the the given name.
	
		def page_named(name)
			my_name = self.class.name.gsub(/Iowa::Application::Content_Classes::/,'').split('::')
			my_name.pop
			my_name << name
			#relative_name = "#{my_name.join('::')}::#{name}"
			relative_name = my_name.join('::')

			iter = 0
			loop do
				if ElementClasses.has_key?(relative_name.downcase)
					break Component.pageNamed(relative_name, @session, self)
				else
					if iter < 1
						iter += 1
						Iowa.app.reloadModified(relative_name,true)
						retry
					else
						break Component.pageNamed(name, @session, self)
					end
				end
			end
		end
		alias :pageNamed :page_named
		alias :componentNamed :page_named
		alias :component_named :page_named


		# These need to be documented.
		def self.proceed_with(meth, page, &block)
			define_method("__#{meth}__primer__".intern, block)
			module_eval(<<-EVAL)
				def #{meth}
					p = page_named('#{page}')
					__#{meth}__primer__(p)
					yield p
				end
			EVAL
		end


		def proceed(meth, page)
			class_eval <<-EVAL
				def #{meth}
					yield page_named('#{page}')
				end
			EVAL
		end

		attr_accessor :session, :prevPage, :subcomponents, :parent

	  # Handle the initial tasks that establish the state of the new object.

		def initialize(name, bindings, attributes, parent)
			@creation_time = Time.now
			@parent = parent
			@session = parent.session if parent

			@subcomponents = {}

			super(name, bindings, attributes)

			awake()
		end

	  # Stub, intended to be overridden in the subclasses, to define behavior
	  # that occurs only once, when the object is first created.

		def awake
		end

	  # Stub, intended to be overridden in the subclasses, to define behavior
	  # that occurs when backtracking.

		def handle_backtrack
		end
		alias :handleBacktrack :handle_backtrack

		# Stub, intended to be overridden in the subclasses, to define behavior
		# that occurs each time the component is displayed.
		
		def setup
		end

		# Takes the result of some method, plus the id of the originating
		# JSON-RP call, and creates the JSON-RPC reponse
		def format_repsonse( result, error, id )
		#File.open('/tmp/jsonid.out','a+') {|fh| fh.puts id.class.name, id.inspect}
		#	id.class.kind_of?(String) ? quote = '"' : quote = ''
		#	File.open('/tmp/jsonid.out','a+') {|fh| fh.puts quote}
		quote = '"'
			'{ "result" : ' + result.to_json + 
				', "error":"' + error.to_s + 
				'", "id": ' +  quote +  id.to_s + quote + ' }'
		end

		def json_methods
			[:list_methods]
		end

		def list_methods
			json_methods
		end

		# Sets up a standard method that will return a URL for handling JSON-RPC
		# requests.
		def json
			session.resource_url do
				# Get the request info
				content = @session.context.request.content
				#File.open('/tmp/jsonin.out','a+') {|fh| fh.puts content}
				
				# Create new JSON lexer
				jsonobj = {}.from_json(JSON::Lexer.new(content))
				# Decode request
				r = nil
				error = nil
				# Verify that it is for an allowed method
				# Call Method
				params = jsonobj['params'] ? jsonobj['params'] : []
				
				begin
					if json_methods.include?(jsonobj[ 'method'].to_sym)
						r = params.size < 1 ?
							self.__send__(jsonobj['method']) :
							self.__send__(jsonobj['method'], *jsonobj['params'])
					end
				rescue Exception => e
					error = e.to_s
				end
				# Return Resource
				Iowa::Resource.new(format_repsonse( r, error, jsonobj['id'] || 'default' ),'text/javascript')
			end
		end

	  # Returns the Application object.

		def application
			Iowa.app
		end
	
	  # Trigger an action on the component.
	
		def invokeAction(action, &b)
			__send__(action, &b)
		end
	
		def handleResponse(context)
			@template = self.class.template
			setup
			handleRequestOrResponse(:handleResponse, context)
		end
	
		def handleRequest(context)
			handleRequestOrResponse(:handleRequest, context)
		end
	
		def handleRequestOrResponse(method, context)
			Iowa.app.reloadModified(self.class.to_s,true)
			
			context.pushRoot self
			@template.__send__(method, context)
			context.popRoot
		end
	
		def getBinding(binding)
			@bindings[binding].get(@parent)
		end
	
		def setBinding(binding, value)
			@bindings[binding].set(@parent, value)
		end
	
		def dup
			copy = super
			@subcomponents.each do |name, child|
				newChild = child.dup
				newChild.parent = copy
				copy.subcomponents[name] = newChild
			end
			copy
		end
	
		def back
			if @prevPage
				yield @prevPage
			end
		end
	
		def reload
		end
	
		def creation_time
			@creation_time
		end
	end

end
